#!/usr/bin/python

from solid import *
from solid.utils import *  # Not required, but the utils module is useful
use("/home/brad/openscad-2014.03/libraries/MCAD/triangles.scad")
use("/home/brad/openscad-2014.03/libraries/MCAD/regular_shapes.scad")
use("/home/brad/openscad-2014.03/libraries/MCAD/gears.scad")

# Source of physical measurements from the patent drawing
import FlyingMachineDrawing as fmd

# NOTE Angles returned by python's trig functions use the traditional
# CCW orientation but openscad's rotate() applies X-axis angles in a 
# clockwise fashion.  There are adjustments throughout the code for this.

TransparentAluminum = [0.77, 0.77, 0.80, 0.01]
Silver = [230/255.0, 232/255.0, 250/255.0,0.9]

defaultColor = Silver
defaultColor = Aluminum 
#defaultColor = TransparentAluminum 
#defaultColor = Transparent

# DONE Make the dampers fit flush to the hull
# DONE Lower the front axle to get the right gap between it and the hull
# DONE Trim the rear axle support from inside the hull
# DONE Figure out where the floating traingle is coming from
# DONE Fix the wrong offset on the trailing edge of the wing
# DONE Fix the wrong offset on the leading edge of the wing
# DONE Fix the width of the forward wing plane
# DONE Get the fin & wing nose to join without a gap
# DONE Fix the wing plane junction
# DONE Add the proper angle to the top edge of the finAngle
# DONE Get the proper angle between the shaft and intake
# DONE Add hangers to the propeller shaft
# DONE add a driveshaft, drive gear and receiver gear 
# DONE Draw proper propellers
# DONE Add the bypass pipes from the bulkhead
# DONE Why is the top front point of the fin a little too far back?
# DONE Why do the gears disappear when rendered?
# TODO Add a plate and cutout on face 3


colorSet1 = {
  'default': defaultColor,
  'debug1' : Blue,
  'debug2' : Red,
  'fin'    : defaultColor,
  'wing'   : defaultColor,
  'intake' : defaultColor,
  'nose'   : defaultColor,
  'center' : defaultColor,
  'tail'   : defaultColor,
  'ruler'  : Oak,
  'razor'  : defaultColor, #Red,
  'tire'   : Black,
  'hub'    : Aluminum,
  'axle'   : BlackPaint,
  'prop_blade' : Oak,
}

colorSet2 = {
  'default': Blue,
  'debug1' : Brass,
  'debug2' : Magenta,
  'fin'    : Green,
  'wing'   : Magenta,
  'intake' : Blue,
  'nose'   : Red,
  'center' : Yellow,
  'tail'   : Green,
  'ruler'  : Iron,
  'razor'  : Oak,
  'tire'   : Black,
  'hub'    : Aluminum,
  'axle'   : BlackPaint,
  'prop_blade' : Oak,
}

class BlankObject(object):
  ''' A blank slate that you can use for storing arbitrary stuff.  Use it like a named tuple. '''
  pass


class sideViewRazor(object):
  def __init__(self):
    ''' Used for slicing the fin and the body '''
    self.width = fmd.wingspan * 1.2
    self.height = fmd.length*4
    self.length = fmd.length*2
    self.a = cube([self.width,self.length,self.height])
    self.a = color(Black)(self.a)
    self.a = left(self.width/2.0)(self.a)    # put it on the centerline

# Each part is created with its "tip" at the origin and aligned
# with the Y axis.  In an aircraft, the Y axis is the direction of
# travel and the origin is the nose.  All other parts are moved back
# from that location.  In OpenSCAD, the "Front" view is LOOKING TO the
# front, not FROM the front.

# Utility to create the various razors and faces we use during construction
def createFacePlane(p1, p2, oldangle, height=fmd.length*20):
  assert p1[0] < p2[0]
  
  run = float(p2[0] - p1[0])
  rise = float(p2[1] - p1[1])
  yOffset = p2[0] - run/2.0 
  zOffset = max(p1[1], p2[1]) - abs(rise/2.0)
  angle = -degrees(atan(rise/run))  # trig functions are opposite direction of openSCAD's rotate()
  length = sqrt(run*run + rise*rise) * 1.2 + 50
  width = fmd.wingspan * 1.2  # make it a little bit wider than the aircraft

  h = abs(height)
  a = cube([width,length,h], center=True)
  
  if height > 0:
    a = up(zOffset+height/2.0)(a)
  else:
    a = down(-zOffset-height/2.0)(a)
    
  a = rotate([angle,0,0])(a)
  a = back(yOffset)(a)

  a = color(defaultColor)(a)
  return a
  
def thin_line(axis, offset, length = 300, line_weight=0.25):
  if axis == 'x':
    line = cube( [ length, line_weight, line_weight], center=True)
  elif axis == 'y':
    line = cube( [ line_weight, length, line_weight], center=True)
  elif axis == 'z':
    line = cube( [ line_weight, line_weight, length], center=True)
    
  return up(offset)(line)
            
def line_through_points(p1x, p1y, p2x, p2y, length = 300, line_weight=0.25):
  rise = p2y - p1y
  run = p2x - p1x
  angle = degrees(atan2(rise, run))
  #print angle
  line = cube( [ length, line_weight, line_weight], center=True)
  line = rotate([0,angle,0])(line)
  line = up(p1y)(line)
  line = left(p1x)(line)
  line += color(Red)(line)
  return line
  
  
class flyingMachine(object):
  # When constructing sub-assemblies, each of them is responsible for creating itself
  # at the origin.  The caller will move it into position.  In other words, each piece 
  # knows its size and shape but the owner knows its orientation and location.

  def __init__(self):
    # Display options
    self.showInternalsOnly = False
    self.useIntakeCutaway =  False
    self.useSideViewCutaway = False
    self.useTopViewCutaway = False
    enableSideViewLines = False
    enableTopViewLines = False

    #enableSideViewLines = True
    #enableTopViewLines = True
    
    
    self.colors = colorSet1
  
    self.lengthNose = fmd.SAF - fmd.TAC
    self.lengthCenter = fmd.TAG - fmd.SAF
    self.lengthTail = fmd.TAH - fmd.TAG

    # These are the diameters at each end of the cylinder that forms the
    # center section.
    self.diameter1 = fmd.TMN * 2
    self.diameter2 = fmd.TMO * 2
    
    # Thickness of the hull
    self.wall = fmd.wallThickness
        
    # TODO - find a better way to create these planes and razors
    
    # face 1 stuff
    angle = -fmd.plane1Angle
    p1 = [fmd.TAC, fmd.SMS]
    p2 = [fmd.SAJ-1, fmd.SMU]
    self.face1Plane = createFacePlane(p1, p2, angle, height=1)
    self.face1Plane = down(self.wall*1.5)(self.face1Plane)   
    self.face1Razor = createFacePlane(p1, p2, angle)
    self.face1Razor = down(self.wall/2.0)(self.face1Razor)

    # face 2 stuff
    angle = 180 - fmd.plane2Angle
    p1 = [fmd.SAJ-1, fmd.SMU]
    p2 = [fmd.SAL, fmd.SMR]
    self.face2Plane = createFacePlane(p1, p2, angle, height=1)
    self.face2Plane = down(self.wall*1.5)(self.face2Plane)   
    self.face2Razor = createFacePlane(p1, p2, angle)
    self.face2Razor = down(self.wall*1.5)(self.face2Razor)

    # TODO There is some serious jury rigging going on here
    # There has to be a cleaner solution
    foo = self.face1Plane * self.face2Plane
    self.face1Plane -= self.face2Razor
    self.face2Plane -= self.face1Razor
    self.face2Plane += foo
    
    # face 3 stuff
    angle = 180 - fmd.plane3Angle
    p1 = [fmd.SAJ, fmd.SMN]
    p2 = [fmd.SAL, fmd.SMR]
    face3offset = 0.7 # works-around a weird artifact when the two faces come to an actual point
    self.face3Plane = createFacePlane(p1, p2, angle, height=-1)
    self.face3Razor = createFacePlane(p1, p2, angle, height=-100)
    self.face3Razor = down(self.wall+face3offset)(self.face3Razor)

    # face4 is the bottom of the center section and doesn't need cutting

    # face 5 stuff
    angle = 180 - fmd.plane5Angle
    p1 = [fmd.SAD, fmd.SMP]
    p2 = [fmd.SAF, fmd.SMO]
    self.face5Plane = createFacePlane(p1, p2, angle, height=-1)
    self.face5Razor = createFacePlane(p1, p2, angle, height=-100)
    self.face5Razor = down(self.wall+0.5)(self.face5Razor)

    self.a = self.assembly() 

    # Side view lines
    if enableSideViewLines:
      lines = union()
      #self.a += line_through_points(fmd.SAJ, fmd.SMN, fmd.SAL, fmd.SMR)
      #self.a += line_through_points(fmd.SAD, fmd.SMP, fmd.SAF, fmd.SMO)
      lines += line_through_points(fmd.SAD, fmd.SMW, fmd.SAD, 0)
      lines += line_through_points(fmd.SAB, fmd.SMW, fmd.SAB, 0)
      lines += line_through_points(fmd.SAC, fmd.SMW, fmd.SAC, 0)
      lines += line_through_points(fmd.SAD, fmd.SMW, fmd.SAD, 0)
      lines += line_through_points(fmd.SAJ, fmd.SMW, fmd.SAJ, 0)
      lines += line_through_points(fmd.SAL, fmd.SMW, fmd.SAL, 0)
      lines += line_through_points(fmd.TAD, fmd.SMW, fmd.TAD, 0)
      lines += line_through_points(fmd.SAF, fmd.SMO, fmd.SAJ, fmd.SMN)
      lines += line_through_points(fmd.SAD, fmd.SMP, fmd.SAF, fmd.SMO)
      lines += line_through_points(fmd.SAD, fmd.SMS, fmd.SAJ, fmd.SMU)
      lines += line_through_points(fmd.SAJ, fmd.SMU, fmd.SAL, fmd.SMR)
      lines += line_through_points(fmd.SAJ, fmd.SMN, fmd.SAL, fmd.SMR)

      lines += line_through_points(fmd.SAG, fmd.SMW, fmd.SAG, 0)
      lines += line_through_points(fmd.SAJ, fmd.SMW, fmd.SAJ, 0)
      lines += line_through_points(fmd.SAJ, fmd.SMW, fmd.SAC, fmd.SMV)
      lines += line_through_points(0, fmd.SMS, fmd.SAL, fmd.SMS)
      lines += line_through_points(0, fmd.SMP, fmd.SAL, fmd.SMP)

      #self.a += line_through_points(fmd.SAD, fmd.SMP, fmd.SAF, fmd.SMO)
      #self.a += line_through_points(fmd.SAD, fmd.SMP, fmd.SAF, fmd.SMO)
      #self.a += line_through_points(fmd.SAD, fmd.SMP, fmd.SAF, fmd.SMO)
      lines = rotate([0,0,90])(lines)
      lines = left(fmd.TMQ+5)(lines)
      self.a += lines

    # Top view lines
    if enableTopViewLines:
      lines = union()
      #lines += rotate([90,0,0])(line_through_points(fmd.TMQ, fmd.TAD, 0, 0))
      #lines += rotate([90,0,0])(line_through_points(fmd.TMQ, fmd.TAD, fmd.TMP, fmd.TAF))
      #lines += rotate([90,0,0])(line_through_points(fmd.TMO, fmd.TAG, fmd.TMP, fmd.TAF))
      lines += line_through_points(fmd.TMQ, fmd.TAD, 0, 0)
      lines += line_through_points(fmd.TMQ, fmd.TAD, fmd.TMP, fmd.TAF)
      lines += line_through_points(fmd.TMO, fmd.TAG, fmd.TMP, fmd.TAF)
      lines += line_through_points(-fmd.TMQ, fmd.TAD, 0, 0)
      lines += line_through_points(-fmd.TMQ, fmd.TAD, -fmd.TMP, fmd.TAF)
      lines += line_through_points(-fmd.TMO, fmd.TAG, -fmd.TMP, fmd.TAF)
      lines = rotate([90,0,0])(lines)
      self.a += lines
    
    #self.a += self.ruler() 

  def color(self, s):
    ''' return an openSCAD color function with specified color from the color map '''
    return color(self.colors[s])
  
  def ruler(self):
    ''' draw a set of reference lines parallel to the Y axis '''
    depth = 0
    ruler  = down(depth)((back(fmd.TAC)(cube([1,fmd.TAC,1]))))
    ruler += down(depth+1)((back(fmd.TAD)(cube([1,fmd.TAD,1]))))
    ruler += down(depth+2)((back(fmd.SAJ)(cube([1,fmd.SAJ,1]))))
    ruler += down(depth+3)((back(fmd.TAG)(cube([1,fmd.TAG,1]))))
    ruler += down(depth+4)((back(fmd.TAH)(cube([1,fmd.TAH,1]))))
    ruler = self.color('ruler')(ruler)
    return ruler
    
    
  def assembly(self):
    ''' Assemble all the major pieces of the aircraft '''
    
    def moveNose(o):
      hullAngle = 0.9   # TODO magic
      o = rotate([hullAngle,0,0])(o)      # Is this a hack or not?  It allows the hull cylinder to align with the reference lines.
      o = back(fmd.TAC+1.4)(o)  # TODO magic
      return up(fmd.SMX)(o)
      
    def moveIntake(o):
      o = back(fmd.TAB)(o)
      return up(fmd.SMX)(o)

    intake = self.intake()
    intake.outside = moveIntake(intake.outside)
    intake.inside  = moveIntake(intake.inside)

    if self.showInternalsOnly:
      # Display just the internal pieces 
      b = intake.outside
      b -= intake.inside
      b += self.bulkhead()
      b += self.propellers()
      b += self.bypassPipes()
      return b

    # The hull has three pieces and an inside and outside.  Put them together and in position
    hull = self.hull()
    
    outside = hull.nose.outside + hull.center.outside + hull.tail.outside
    outside = moveNose(outside)
    
    #return outside+self.fin()

    inside = hull.nose.inside + hull.center.inside + hull.tail.inside
    inside = moveNose(inside)

    # Each of the flat faces are the intersection of a plane and the solid body hull
    plate1 = outside * self.face1Plane
    plate2 = outside * self.face2Plane
    plate3 = outside * self.face3Plane
    plate5 = outside * self.face5Plane

    a = outside
  
    # Now that we have the plates constructed, we can start carving away on the solid body to get
    # the rough outline of the craft
    a -= self.face1Razor
    a -= self.face2Razor
    a -= self.face3Razor
    a -= self.face5Razor
    outside_trimmed = a    # save this so we can use it again later
    a -= inside

    # Add the fin and wing
    # Subtracting out the razors on the fin and wing at the same time gives the proper point to the nose
    finAndWing = self.fin() + self.wing()
    finAndWing = finAndWing - self.leadingEdgeRazor - self.finRazor

    # Subtracting out the trimmed solid body prevents the wing and fin from extending inside
    finAndWing -= outside_trimmed
    a += finAndWing 

    # Add the flat plates
    a += plate1
    a += plate2
    a += plate5

    # The flat plate on face 3 is experimental
    if False:
      a += plate3
      face3Cutout = cylinder(r=self.diameter1*0.4, h=10)
      face3Cutout = scale([1,2,1])(face3Cutout)
      face3Cutout = rotate([-22,0,0])(face3Cutout)
      face3Cutout = back((fmd.SAJ+fmd.SAL)/2.0+5)(face3Cutout)
      face3Cutout = up(5)(face3Cutout)
      a -= face3Cutout


    # Insert the internal bulkhead between compartments
    bulkhead = self.bulkhead() * outside_trimmed
    a += bulkhead

    # Insert the intake pipe and its hollowness
    intake = self.intake()
    intake.outside = moveIntake(intake.outside)
    intake.inside  = moveIntake(intake.inside)
    a += intake.outside
    a -= intake.inside

    a += self.bypassPipes()

    # Cut out the inlets on the upper side of the hull
    a -= self.upperInlets()

    # Make a hole for the dampers and put them in
    cutout,dampers = self.dampers()
    a = a - cutout + dampers   

    # Insert the propellers and wheel
    a += self.propellers()
    a += self.wheels()
    
    # Set to True to show a side-view cutaway like in the patent drawing
    if self.useSideViewCutaway:
      rzr = cube([200,200,200],center=True)
      rzr = left(100)(rzr)
      rzr = back(100)(rzr)
      rzr = color(Black)(rzr)
      a -= rzr
      
    # Set to True to show a top-view cutaway like in the patent drawing
    if self.useTopViewCutaway:
      rzr = cube([200,200,200],center=True)
      rzr = up(100+fmd.SMX+2)(rzr)
      rzr = back(100)(rzr)
      rzr = color(Black)(rzr)
      a -= rzr
      
    return a

  def bypassPipes(self):
    length = 12
    curvature = fmd.bypassOD * 0.8
    b = torus2(curvature,1.2)
    rzr = cube([20,20,20],center=True)
    rzr = forward(10)(rzr)
    b -= rzr
    
    b = left(curvature)(b)
    rzr = rotate([0,0,20])(rzr)
    rzr = left(2)(rzr)
    b -= rzr
    b = rotate([0,110,0])(b)
    b = back(length)(b)

    a = cylinder(h=length, r=fmd.bypassOD/2.0)
    a = rotate([90,0,0])(a)

    a += b
    
    angle = 180 - fmd.plane4Angle
    a = rotate([angle,0,0])(a)
    a = back(fmd.SAH-1.5)(a)
    a = left(4)(a)
    a = up(fmd.SMN+fmd.bypassOD+1.1)(a)
    
    a = a + mirror([1,0,0])(a)
    return a
    
    
  def fin(self):
    # The gross dimensions of the fin run from the top of the intake pipe to SMW
    # and from the reference line to the peak of the fuselage.
    finHeight = fmd.SMW - fmd.SMS
    finLength = fmd.SAJ
    f = cube([self.wall, finLength, finHeight]) 
    f = back(finLength)(f)
    f = left(self.wall/2.0)(f)  # put it on the centerline
    f = self.color('fin')(f)

    # create the front top angle
    razorObject = sideViewRazor()
    razor = razorObject.a
    angle = fmd.finAngle1
    razor = rotate([angle,0,0])(razor)
    razor = up(fmd.SMT)(razor)
    razor = self.color('razor')(razor)
    self.finRazor = razor  # save it off so we can apply it later

    # create the slight taper on the very top
    razorObject = sideViewRazor()
    razor = razorObject.a
    rise = fmd.SMW - fmd.SMV
    run = fmd.SAJ
    angle = -degrees(atan(rise/run))
    razor = rotate([angle,0,0])(razor)
    razor = up(fmd.SMW-fmd.SMS)(razor)
    razor = back(fmd.SAJ)(razor)
    razor = self.color('razor')(razor)
    f = f - razor
    
    # create the front bottom angle
    razorObject = sideViewRazor()
    razor = razorObject.a
    razor = rotate([0,180,0])(razor)
    razor = back(fmd.TAC)(razor)
    razor = up(fmd.SMT-fmd.SMS)(razor)
    angle = 180 - fmd.finAngle2
    razor = rotate([angle,0,0])(razor)
    razor = self.color('razor')(razor)
    f = f - razor
    
    f = up(fmd.SMS)(f)

    return f
  
  def wing(self):
    thickness = 1

    # Create one side of the front, upwardly angled piece
    # This starts as a flat plane in the X-Y axis then gets trimmed to create the sweep of the 
    # leading edge.  Then it gets rotated around the X axis to create the upward rise toward the nose.
    x = fmd.TMQ
    y = fmd.TAD 

    theta = degrees(atan(float(x)/float(y)))  # wing sweep in the X-Y plane
    
    alpha = 180 - fmd.finAngle2	      # upward inclination in the Y-Z plane
    ralpha = radians(alpha)
    
    horizontalOverlap = 0.5  # magic
    horizontalOffset = cos(ralpha)*fmd.TAD + horizontalOverlap
    
    z = fmd.SMT - fmd.SMX
    length = sqrt(z*z+y*y)   # hypotonuese in Y-Z plane

    a = cube([fmd.TMQ, length, thickness])
    a = rotate([alpha,0,0])(a)               # tilt the wing up
    a = back(y)(a)                           # move it back to meet the center portion
    offset = fmd.TAD * sin(ralpha)
    a = down(offset)(a)  

    a = a + mirror([1,0,0])(a)

    # Build the razor that we will need later to cut the sweep on the leading edge
    rzrWidth = 20
    rzrLength = 40
    rzrHeight = (fmd.SMT - fmd.SMX) * 3
    rzr = cube([rzrWidth, rzrLength+5, rzrHeight])
    rzr = down(rzrHeight/2.0)(rzr)
    rzr = back(rzrLength)(rzr)
    rzr = rotate([0,0,theta])(rzr)  # create the swept-back leading edge
    rzr = up(fmd.SMT)(rzr)
    rzr = self.color('razor')(rzr)
    self.leadingEdgeRazor = rzr + mirror([1,0,0])(rzr)

    #
    # Create the center section of the wing
    #
    
    # TODO I think the sequence of construction can be simplified 
    
    # Compute the angle that the wing makes as it tapers toward the back
    x = float(fmd.TMQ - fmd.TMP)
    y = float(fmd.TAF - fmd.TAD)
    angle = degrees(atan(x/y))
    
    b = cube([fmd.TMQ, fmd.TAH-fmd.TAD, thickness])
    rzr = cube([fmd.TMQ+20, fmd.TAG-fmd.TAD+20, thickness+20])
    rzr = down(1)(rzr)
    rzr = self.color('razor')(rzr)
    rzr = rotate([0,0,angle])(rzr)
    rzr = right(fmd.TMQ)(rzr)
    b = b - rzr

    # Create the trailing edge of the wing
    x = float(fmd.TMP - fmd.TMO)
    y = float(fmd.TAG - fmd.TAF)
    angle = degrees(atan(y/x))
    
    rzr = cube([fmd.TMQ+20, fmd.TAG-fmd.TAD+20, thickness+20])
    rzr = down(2)(rzr)
    rzr = self.color('razor')(rzr)
    rzr = rotate([0,180,0])(rzr)
    rzr = rotate([0,0,-angle])(rzr)
    rzr = forward(fmd.TAF - fmd.TAD-2)(rzr)
    rzr = right(fmd.TMQ)(rzr)
    b = b - rzr

    # Create a mirror image and combine the two
    b = b + mirror([1,0,0])(b)
  
    # Line it up with the front section
    b = rotate([0,0,180])(b)  # flip it around to the desired orientation
    b = back(fmd.TAD)(b)
    b = down(offset)(b)
            
    # Put the parts together, position it and set the color
    w = a + b
    w = up(fmd.SMT)(w)
    w = self.color('wing')(w)
    return w
  
  def oneBlade(self):
    ''' Draw a single propeller blade '''
    # Start with a cyliner, cut half of it off and then slice away all but an angle.
    a = cylinder(h=0.2, r=8.0)
    rzr = cube([20,20,20],center=True)
    rzr = left(10)(rzr)
    a -= rzr
    rzr = rotate([0,0,35])(rzr)
    a *= rzr
    
    a = rotate([-10,30,0])(a)   # A blade without pitch doesn't move any air
    
    a = up(.5)(a)
    a = self.color('prop_blade')(a)
    return a
    
  def driveshaft(self):
    shaftLength = 3
    driveGear = gear(50, circular_pitch=12)
    driveGear = linear_extrude(height = 1, center = True, convexity = 10)(driveGear)
    driveGear = up(2.2)(driveGear)
    
    shaft = cylinder(r=0.5, h=shaftLength)
    
    a = shaft + driveGear
    a = rotate([90,0,0])(a)
    a = translate([0,-fmd.SAH+1.2,(fmd.SMP-fmd.SMO)/2.0+11])(a)

    return a

  def propellers(self):
    shaftLength = fmd.SAK - fmd.SAE
    
    b1pos = 0.1
    b2pos = (fmd.SAH-fmd.SAE)*56.0/94.0
    b3pos = fmd.SAH-fmd.SAE-1.1
    b4pos = fmd.SAJ-fmd.SAE
    self.gearPos = fmd.SAH-fmd.SAE+1  # gets used later for the drive gear
    
    bearing = cylinder(r=0.7,h=1)
    
    bearing1 = up(b1pos)(bearing)
    bracket1 = cylinder(r=0.3, h=1)
    bracket1 = rotate([90,0,0])(bracket1)
    bracket1 = up(0.6)(bracket1)
    bracket1 = forward(1.5)(bracket1)
    bearing1 += bracket1
    
    bearing2 = up(b2pos)(bearing)
    bracket2 = cylinder(r=0.3, h=3.5)
    bracket2 = rotate([90,0,0])(bracket2)
    bracket2 = up(b2pos+0.5)(bracket2)
    bracket2 = forward(1.8)(bracket2)
    bearing2 += bracket2

    bearing3 = up(b3pos)(bearing)
    bracket3 = cylinder(r=0.3, h=1)
    bracket3 = rotate([90,0,0])(bracket3)
    bracket3 = up(b3pos+0.5)(bracket3)
    bracket3 = back(0.5)(bracket3)
    bearing3 += bracket3
    
    bearing4 = up(b4pos)(bearing)
    bracket4 = cylinder(r=0.3, h=22.5)
    bracket4 = rotate([90,0,0])(bracket4)
    bracket4 = up(b4pos+0.5)(bracket4)
    bracket4 = forward(9.5)(bracket4)
    bearing4 += bracket4
    
    gear1 = gear(20, circular_pitch=15)
    gear1 = linear_extrude(height = 1, center = True, convexity = 10)(gear1)
    gear1 = up(self.gearPos)(gear1)
    
    shaft = cylinder(r=0.5, h=shaftLength)
    hub = color(Stainless)(cylinder(r=1.5,h=1))
    blade = self.oneBlade()
    smallBlade = scale([1,.85,1])(blade)

    prop = hub + blade + rotate([0,0,90])(blade) + rotate([0,0,180])(blade) + rotate([0,0,270])(blade)
    smallProp = hub + smallBlade + rotate([0,0,180])(smallBlade)

    prop1 = up(fmd.SAI-fmd.SAE)(rotate([0,0,-30])(prop))
    prop2 = up(shaftLength-1.1)(rotate([0,0,30])(smallProp))
    #prop1 = up(fmd.SAI-fmd.SAE)(rotate([0,0,-25])(prop))
    #prop2 = up(shaftLength-1.1)(rotate([0,0,-25])(smallProp))
    a = bearing1 + bearing2 + bearing3 + bearing4 + gear1 + shaft + prop1 + prop2
    a = rotate([90,0,0])(a)
    a = translate([0,-fmd.SAE,fmd.SMQ])(a)
    return a + self.driveshaft()
  
  def dampers(self):
    ''' Returns a hole that fits the dampers and the dampers themselves '''
    # There are no details in the drawing about this so I had to make it all up
    r1 = 2
    r2 = r1 - 0.2
    height = 0.7
    
    def move(o):
      o = rotate([-90,-43,-3])(o)  # magic that allow the dampers to lay flat(ish) to the hull
      o = left(fmd.FMN*0.7)(o)
      o = back(fmd.SAG)(o)
      o = up(fmd.SMO+3)(o)
      o = o + mirror([1,0,0])(o)
      return o
      
    a = cylinder(r=r1,h=height,center=True)
    a = rotate([0,90,0])(a)

    c = cylinder(r=r2,h=height*3,center=True)
    c = rotate([0,90,0])(c)
    
    b = cube([0.2,10,1.0], center=True)
    b = rotate([0,20,0])(b)

    z = 0.8
    slats = b
    slats += up(z)(b) + up(2*z)(b) + up(3*z)(b)
    slats += down(z)(b) + down(2*z)(b) + down(3*z)(b)
    slats = slats * a
    a = a - c + slats

    c = move(c)   # c is the cutout
    a = move(a)   # a is the circular frame plus the slats
    
    c = self.color('default')(c)
    a = self.color('default')(a)
    return c,a
    
  def upperInlets(self):
    ''' Holes on the upper side of the fuselage near the propellers '''
    length = fmd.upperInletLength
    width = fmd.upperInletWidth
    height = self.wall + 3   # long enough to cut through the hull plus a little
    angle = 38   # magic number
    xOffset = fmd.TMN * sin(radians(angle)) + width/2.0
    yOffset = fmd.SAI+length/2.0    # centered on the end of the fin
    zOffset = (fmd.SMT+fmd.SMU)/2.0  # halfway between T and U
    
    a = cube([width, length, height], center=True)
    a = rotate([0,angle,0])(a)
    a = translate([xOffset, -yOffset, zOffset])(a)
    a = a + mirror([1,0,0])(a)
    a = self.color('default')(a)
    return a
  
  def bulkhead(self):
    ''' Internal wall through which the pipe and propeller shaft passes '''
    a = cube([fmd.TMO*2, 1, fmd.SMW])
    a = translate([-fmd.TMO,-fmd.SAH+1,0])(a)
    a = self.color('default')(a)
    return a
    
  
  # The "generally ovoid" fuselage has three constituent parts.  These
  # are the nose, the center section and the tail section.
  def hull(self):  
    a = BlankObject()
    a.nose = self.nose()

    a.center = self.centerSection()
    a.center.outside = back(self.lengthNose)(a.center.outside)
    a.center.inside  = back(self.lengthNose)(a.center.inside)

    a.tail = self.tail()
    a.tail.outside = back(self.lengthNose+self.lengthCenter)(a.tail.outside)
    a.tail.inside  = back(self.lengthNose+self.lengthCenter)(a.tail.inside)

    return a


  def nose(self):
    ''' From where the intake enters the body to where the body becomes straight '''
    length = self.lengthNose
    r = self.diameter1 / 2.0
    scaleFactor = [1,2,1]

    a = BlankObject()
    
    a.outside = scale(scaleFactor)(sphere(r))
    a.inside = scale(scaleFactor)(sphere(r-self.wall))
    razor = cube([50,50,50], center=True)
    razor = back(25)(razor)
      
    a.outside = a.outside - razor
    a.outside = back(self.diameter1)(a.outside) # Put the tip at the origin
    a.inside = back(self.diameter1)(a.inside)
    a.outside = self.color('nose')(a.outside)
    a.inside = self.color('nose')(a.inside)
    return a

  def centerSection(self):
    ''' A tapered cylinder '''
    length = self.lengthCenter
    r1 = self.diameter1 / 2.0
    r2 = self.diameter2 / 2.0
    
    a = BlankObject()
        
    def prep(o):
      o = rotate([90,0,0])(o)
      o = self.color('center')(o)
      return o
      
    a.outside = prep(cylinder(h=length, r1=r1, r2=r2))
    a.inside = prep(cylinder(h=length, r1=r1-self.wall, r2=r2-self.wall))
    
    return a

  def tail(self):
    length = self.lengthTail
    r1 = self.diameter2 / 2.0
    r2 = r1 - self.wall

    a = BlankObject()

    # Using the razor leaves a thin artifact between the hemispheres.  Can do without it
    #rzr = cube([r1*2,r1*2,r1*2])
    #rzr = left(r1)(rzr)
    #rzr = down(r1)(rzr)

    a.outside = sphere(r1) #- rzr
    a.inside = sphere(r2) #- rzr

    a.outside = self.color('tail')(a.outside)
    a.inside  = self.color('tail')(a.inside)
    return a
    
  def intake(self):
    length = fmd.SAH - fmd.SAB
    r1 = fmd.intakeOD / 2.0
    wall = (fmd.intakeOD - fmd.intakeID) / 2.0

    a = BlankObject()

    extraLength = 1.0
    a.outside = cylinder(h=length, r=r1);
    a.inside = cylinder(h=length+extraLength, r=r1-wall)
    
    if self.useIntakeCutaway:
      rzrLength = length+extraLength
      rzr = cube([5,5,rzrLength], center=True)
      rzr = left(2.5)(rzr)
      rzr = up(length/2.0)(rzr)
      a.outside -= rzr

    angle = -fmd.intakeAngle
    a.outside = rotate([90+angle,0,0])(a.outside)
    a.inside  = rotate([90+angle,0,0])(a.inside)
    a.inside = forward(extraLength/2.0)(a.inside)  # stick out the front and back to ensure a clean cutout
    a.inside = self.color('intake')(a.inside)
    a.outside = self.color('intake')(a.outside)
        
    return a

  def wheel(self):
    r1 = fmd.wheelOD/2.0
    r2 = fmd.wheelID/2.0

    rubberThickness = 0.5
    rubber = torus2(r1-rubberThickness, rubberThickness)
    rubber = self.color('tire')(rubber)

    hub = cylinder(h=.95,r=r1-.3)
    hub = down(0.475)(hub)
    hub += torus2(r1/3.0, rubberThickness)   # add a ring in the middle for decoration
    hub = self.color('hub')(hub)
    
    axle = cylinder(h=2.75, r=0.4) 
    axle = down(0.75)(axle)    
    axle = self.color('axle')(axle)
    
    a = rubber + hub + axle
    a = rotate([0,90,0])(a)

    a = up(r1)(a)
    return a
    
  # The carriage has three parts
  #	upright attaches the main beam to the fuselage
  #	mainbeam is the horizontal member
  #	legs are the things that attach the mainbeam to the axles
  def upright(self):
    # NOTE This function is full of magic numbers
    # NOTE This function is full of magic numbers
    
    radius = 1.0
    thickness = 1.0

    a = cube([3,1.5,1],center=True)
    a = up(0.5)(a)
    #a = color(Blue)(a)
    rzr = cylinder(r=radius, h=2)
    rzr = scale([1,5,1])(rzr)
    rzr = left(1.5)(rzr)
    rzr = forward(4)(rzr)
    rzr = down(0.5)(rzr)
    rzr = rzr + mirror([1,0,0])(rzr)
    #rzr = color(Red)(rzr)
    a -= rzr
    a = rotate([90,0,0])(a)
    
    a = forward(0.5)(a)
    a = up(4.2)(a)
    
    a = self.color('axle')(a)
    return a
 
  
  def legs(self):
    # NOTE This function is full of magic numbers
    
    radius = 2.0
    thickness = 1.0
    fillerHeight = .5
    fillerWidth = 0.6
    arcstart = 90
    arcfinish = 165
    offset = self.mainbeam_length / 2.0

    a = cylinder(r=radius, h=1)
    b = cylinder(r=radius-0.5, h=5)
    b = down(1)(b)
    b = scale([0.7,1,1])(b)
    a -= b

    a = rotate([90,0,0])(a)
    a = scale([1,1,2])(a)
    
    rzr = cube(6,center=True);
    rzr = down(1.5)(rzr)
    a -= rzr

    rzr2 = cube(5,center=True)
    rzr2 = right(2.5)(rzr2)
    rzr2 = up(3)(rzr2)
    a -= rzr2
       
    a = up(0.1)(a)
    a = left(offset)(a)
    a = forward(0.5)(a)
    
    a = a + mirror([1,0,0])(a)
    a = self.color('axle')(a)
    
    return a
    
  def mainbeam(self):
    topOfBeam = fmd.FAC
    width = 1.0
    offset = topOfBeam - width / 2.0
    a = cube([self.mainbeam_length,width,width],center=True)
    a = up(offset)(a)
    a = self.color('axle')(a)
    return a
        
  def carriage(self):
    self.mainbeam_length = 2*(fmd.FMN-2.6)

    a = self.mainbeam()
    b = self.legs()
    c = self.upright()

    return a + b + c

  def wheelsAndAxles(self):
    lw = self.wheel()
    lw = left(fmd.FMN)(lw)
    rw = mirror([1,0,0])(lw)
    a = lw + rw + self.carriage()
    return a
        
  def frontWheels(self):
    a = self.wheelsAndAxles()
    swivel = cylinder(r=1.0,h=2.0)
    swivel = up(4.5)(swivel)
    
    a += swivel
    a = back(fmd.SAG)(a)
    return a
    
  def backWheels(self):
    a = self.wheelsAndAxles()
    a = back(fmd.SAJ)(a)
    return a
    
  def wheels(self):
    #return self.frontWheels()
    #return self.backWheels()
    return self.frontWheels() + self.backWheels()
  

if __name__ == '__main__':
  SEGMENTS = 50 #120
  f = flyingMachine()
  file_out = "FlyingMachine.scad"
  scad_render_to_file( f.a, file_out, file_header='$fn = %s;'%SEGMENTS, include_orig_code=True)  
  
