''' 
Contains key dimensions I measured from the drawings on Patent 922,972.
These were measured in mm but no units were specified in the drawings.
There are three perspectives in the patent - side, top and front.
On each perspective, I drew straight lines going through key points
and labeled them with letters.  The constants in this module are the 
distances between the lines.  Examples:

SAB - Side view perspective, distance from line A to line B
TMP - Top view, distance from M to P
FAD - Front view, distance from A to D

Where there are two measurements that should be the same but aren't quite the
same in the drawing, I picked one and commented out the (apparently) bad one.

'''

from math import degrees, atan, atan2, sin, cos, radians, degrees

def angleInDegrees(rise, run):
  ''' Simple geometric utility '''
  return degrees(atan2(rise, run))
  

# Lengthwise dimensions in side-view.  A is the forward tip of the fin.
SAB = 9
SAC = 11.9
SAD = 21.3
SAE = 28
SAF = 39 
SAG = 40.3
SAH = 75.5
SAI = 82
SAJ = 86.9
SAK = 95.9
SAL = 125
SAM = 22.6
SAN = 87.7
SAO = 92.3

# Vertical dimensions in side-view.  M is the ground
SMN = 4.4 
SMO = 6.7 
SMP = 15 
SMQ = 17.7
SMR = 19.8
SMS = 19.1
SMT = 22.6
SMU = 27.9
SMV = 34.3  # I believe he intended the top of the fin to be horizontal but the drawing is just slightly off
SMW = 35.9 
SMX = (SMP + SMS) / 2.0
SMY = 8.4
SMZ = 26.8
SMA = 4.1

# Lengthwise dimensions in top-view.  A is the forward tip of the fin.
# Some of these measurements should exactly match corresponding measurements in the side-view
TAB = SAB  
TAC = 18   
TAD = SAE  
TAE = SAH  
TAF = 106
TAG = 111.4
TAH = SAL

# Horizontal dimensions in top-view.  M is the centerline
TMN = 10.5
TMO = 12.3
TMP = 23.4
TMQ = 27

# Vertical dimensions in forward-view.  A is the ground
FAC = SMA
FAB = FAC / 2.0
FAD = SMO
FAE = SMQ
FAF = SMT
#FAG = 27.5
#FAH = SMU
#FAI = SMW

# Horizontal dimensions in forward-view.  M is the centerline
FMN = 10.5 
FMO = TMQ

# Miscellaneous measurements
wallThickness = 1
intakeWallThickness = 1

intakeOD = SMS - SMP
intakeID = intakeOD - intakeWallThickness

intakeAngle = degrees(atan(intakeWallThickness/(SAH - SAB)))

bypassOD = intakeOD * 0.6
bypassID = bypassOD - 0.7

wheelOD = SMA
wheelID = wheelOD - 0.9
wheelWidth = 1
wheelTrackWidth = FMN * 2
wheelbase = SAJ - SAG

# Upper edge of fin
finAngle1 = angleInDegrees((SMV - SMT), SAC)

# Lower edge of fin
finAngle2 = angleInDegrees((SMT - SMS), -TAC)

# It appears that the upward inclination of the wings is the same as finAngle2.
# The drawing shows the wings and the lower edge of the fin in alignment.
wingAngle = finAngle2

wingspan = TMQ * 2
length = SAL
height = SMW

upperInletLength = 10
upperInletWidth = 3

# In the side-view, the top & bottom of the fuselage are flattened.  Roughly speaking
# the body looks like an irregular five-sided polyhedron.  In order to create these
# flat surfaces, we have to know the angles that each of them makes with the horizontal.
# The angles are measured counter-clockwise relative to the X axis.
# The flat surfaces are:
# plane1 - forward face running from the intake to the peak of the fuselage
# plane2 - rearward face running from the peak to the tip of the tail
# plane3 - rearward face running from the tip of the tail to the rear landing gear
# plane4 - bottom portion of the fuselage
# plane5 - forward face running from the front landing gear to the intake

plane1Angle = angleInDegrees((SMU - SMS), (SAJ - TAC))  # TODO Should TAC be SAD?
plane2Angle = angleInDegrees((SMU - SMR), (SAJ - SAL))
plane3Angle = angleInDegrees((SMR - SMN), (SAL - SAJ))
plane4Angle = angleInDegrees((SMO - SMN), (SAF - SAJ))
plane5Angle = angleInDegrees((SMP - SMO), (TAC - SAF))

#
# main - Verify some relationships and print some key values
# 
if __name__ == '__main__':
  #assert SAD == TAC 
  #assert TMO == FMN
  
  assert TMN == TMN
  assert (FAH - FAD) == (2 * TMN)

  assert intakeOD > intakeID

  print "finAngle1 =", finAngle1
  print "finAngle2 =", finAngle2
  print "plane1Angle = ", plane1Angle
  print "plane2Angle = ", plane2Angle
  print "plane3Angle = ", plane3Angle
  print "plane4Angle = ", plane4Angle
  print "plane5Angle = ", plane5Angle
  